/* Copyright (C) 2010 Stefan Frenz
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */

package ui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Enumeration;
import java.util.Vector;

import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.SwingUtilities;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;

import osio.sun.SunOS;

import compbase.Context;

/**
 * SCG: manage a compilation with a GUI
 * 
 * @author S. Frenz
 * @version 100429 added option for "interface parents array" (-j), removed "structs as objects" (-r)
 *  version 100424 removed options for "return missing" (-R), added check of screen size
 *  version 100422 initial version
 */

public class SCG extends JFrame {
  private static final long serialVersionUID = 1L;

  public static void main(String argv[]) {
    SwingUtilities.invokeLater(new Runnable() {
      public void run() {
        new SCG();
      }
    });
	}
  
  private static class LabelComp extends JPanel {
    private static final long serialVersionUID = 1L;
    public LabelComp(String text, JComponent comp) {
      setLayout(new BorderLayout(2, 0));
      add(new JLabel(text), BorderLayout.WEST);
      add(comp, BorderLayout.CENTER);
    }
  }
  
  private static class ThreeComp extends JPanel {
    private static final long serialVersionUID = 1L;
    public ThreeComp(JComponent center, JComponent southWest, JComponent southEast) {
      setLayout(new BorderLayout(2, 0));
      add(center, BorderLayout.CENTER);
      JPanel south=new JPanel(new GridLayout(1, 1, 2, 0));
      south.add(southWest);
      south.add(southEast);
      add(south, BorderLayout.SOUTH);
    }
  }
  
  private JTextField result;
  private JButton run;
  private JFileChooser fileChooser;
  private Vector<String> args;
  
  private JCheckBox v_verbose, V_verboseTiming;
  private JCheckBox L_flat, l_streamline;
  private JCheckBox M_indirect, m_dynaMem;
  private JTextField i_options, H_listFiles;
  private JCheckBox h_heapAssign, c_allAssign;
  private JCheckBox w_alternateNew, x_stackExtreme;
  private JCheckBox C_skipArrayCheck, B_skipBoundCheck;
  private JCheckBox b_rteBound, n_rteNull;
  private JCheckBox X_explicitConv, d_mthdDebug;
  private JCheckBox g_genAllUnits, G_genAllMthds;
  private JCheckBox j_genIntfParents;
  private JCheckBox q_relManager, Q_srcHint;
  private JCheckBox W_printCode;
  private JCheckBox F_profiler, f_throwWithoutCatch;
  private JCheckBox y_byteString, k_skipInlineMthd;
  private JCheckBox Y_nativeFloatInComp;
  private JTextField N_startMthd;
  private JTextField I_maxInlLev, S_autoInlStmts;
  private JTextField K_allocMask, s_imgSize;
  private JComboBox t_arch;
  private JTextField T_archOpt;
  private JTextField a_addr, Z_reloc;
  private JTextField e_embAddr;
  private JCheckBox E_constInRam;
  private JTextField P_specialHeader, p_namePrefix;
  private JComboBox o_out;
  private JTextField O_outPar;
  private JComboBox u_sym;
  private JTextField U_symPar;
  private JTextField D_code, D_gcc, D_mthd, D_sym, D_rel;
  private JList files, dirs, dirsNonRec;
  private DefaultListModel filesList, dirsList, dirsNonRecList;
  private JButton fileAdd, fileDel, dirAdd, dirDel, dirNRAdd, dirNRDel;
  
  private ActionListener defaultOptionChangedListener=new ActionListener() {
    private static final long serialVersionUID = 1L;
    public void actionPerformed(ActionEvent e) {
      triggerEvaluation(e.getSource());
    }
  };
  private DocumentListener defaultTextChangedListener=new DocumentListener() {
    public void removeUpdate(DocumentEvent e) {
      triggerEvaluation(e.getDocument());
    }
    public void insertUpdate(DocumentEvent e) {
      triggerEvaluation(e.getDocument());
    }
    public void changedUpdate(DocumentEvent e) {
    }
  };
  private ActionListener defaultAddDelListener=new ActionListener() {
    private static final long serialVersionUID = 1L;
    public void actionPerformed(ActionEvent e) {
      triggerAddDel(e.getSource());
    }
  };
  private ActionListener runListener=new ActionListener() {
    private static final long serialVersionUID = 1L;
    public void actionPerformed(ActionEvent e) {
      doCompile();
    }
  };
  
  private SCG() {
    super("smfCompiler option selector GUI");
    setDefaultCloseOperation(EXIT_ON_CLOSE);
    setLayout(new BorderLayout());
    JPanel tmpPanel=new JPanel();
    tmpPanel.setLayout(new GridLayout(0, 4, 2, 2));
    tmpPanel.add(L_flat=new JCheckBox("flat memory"));
    L_flat.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(l_streamline=new JCheckBox("streamline objects"));
    l_streamline.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(M_indirect=new JCheckBox("indirect scalars"));
    M_indirect.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(m_dynaMem=new JCheckBox("movable objects"));
    m_dynaMem.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(v_verbose=new JCheckBox("verbose"));
    v_verbose.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(V_verboseTiming=new JCheckBox("verbose timing"));
    V_verboseTiming.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(new LabelComp("options from file", i_options=new JTextField()));
    i_options.getDocument().addDocumentListener(defaultTextChangedListener);
    tmpPanel.add(new LabelComp("log compiled files to file", H_listFiles=new JTextField()));
    H_listFiles.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(h_heapAssign=new JCheckBox("rte-call for heap pointer assignments"));
    h_heapAssign.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(c_allAssign=new JCheckBox("rte-call for all pointer assignments"));
    c_allAssign.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(w_alternateNew=new JCheckBox("alternate object layout"));
    w_alternateNew.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(x_stackExtreme=new JCheckBox("stack extreme check"));
    x_stackExtreme.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(C_skipArrayCheck=new JCheckBox("skip array check"));
    C_skipArrayCheck.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(B_skipBoundCheck=new JCheckBox("skip bound check"));
    B_skipBoundCheck.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(b_rteBound=new JCheckBox("rte-call for bound exception"));
    b_rteBound.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(n_rteNull=new JCheckBox("rte-call for null exception"));
    n_rteNull.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(X_explicitConv=new JCheckBox("require explicit conversion"));
    X_explicitConv.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(d_mthdDebug=new JCheckBox("create method binary debug files"));
    d_mthdDebug.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(g_genAllUnits=new JCheckBox("create all units (don't skip)"));
    g_genAllUnits.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(G_genAllMthds=new JCheckBox("create all methods (don't skip)"));
    G_genAllMthds.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(j_genIntfParents=new JCheckBox("generate interface parents array"));
    j_genIntfParents.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(q_relManager=new JCheckBox("enable relation manager"));
    q_relManager.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(Q_srcHint=new JCheckBox("integrate source hints"));
    Q_srcHint.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(W_printCode=new JCheckBox("print code of methods"));
    W_printCode.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(F_profiler=new JCheckBox("enable profiler call for every method"));
    F_profiler.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(f_throwWithoutCatch=new JCheckBox("disable throw-frames (no catch)"));
    f_throwWithoutCatch.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(y_byteString=new JCheckBox("Strings with byte-value (not char)"));
    y_byteString.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(k_skipInlineMthd=new JCheckBox("skip method objects for inline methods"));
    k_skipInlineMthd.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(Y_nativeFloatInComp=new JCheckBox("prefer native float code in compiler"));
    Y_nativeFloatInComp.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(new LabelComp("start code in method", N_startMthd=new JTextField()));
    N_startMthd.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(new LabelComp("maximum inline level", I_maxInlLev=new JTextField()));
    I_maxInlLev.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(new LabelComp("max. stmt. count for auto-inline", S_autoInlStmts=new JTextField()));
    S_autoInlStmts.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(new LabelComp("align-mask for allocated block", K_allocMask=new JTextField()));
    K_allocMask.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(new LabelComp("memory image size", s_imgSize=new JTextField()));
    s_imgSize.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(new LabelComp("architecture", t_arch=new JComboBox(new Object[] {"ia32", "ia32opt", "amd64", "atmega", "atmegaOpt", "ssa"})));
    t_arch.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(new LabelComp("architecture options", T_archOpt=new JTextField()));
    T_archOpt.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(new LabelComp("address", a_addr=new JTextField()));
    a_addr.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(new LabelComp("relocate by GB", Z_reloc=new JTextField()));
    Z_reloc.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(new LabelComp("embedded RAM address", e_embAddr=new JTextField()));
    e_embAddr.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(E_constInRam=new JCheckBox("constant objects in RAM"));
    E_constInRam.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(new LabelComp("image header (or \"none\")", P_specialHeader=new JTextField()));
    P_specialHeader.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(new LabelComp("output name prefix", p_namePrefix=new JTextField()));
    p_namePrefix.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(new LabelComp("output format", o_out=new JComboBox(new Object[] {"raw", "boot"})));
    o_out.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(new LabelComp("output options", O_outPar=new JTextField()));
    O_outPar.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(new LabelComp("in-system symbol generator", u_sym=new JComboBox(new Object[] {"none", "raw", "rte", "mthd", "reflect"})));
    u_sym.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(new LabelComp("symbol generator options", U_symPar=new JTextField()));
    U_symPar.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(new LabelComp("debug writer \"code\" file", D_code=new JTextField()));
    D_code.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(new LabelComp("debug writer \"gcc\" file", D_gcc=new JTextField()));
    D_gcc.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(new LabelComp("debug writer \"mthd\" file", D_mthd=new JTextField()));
    D_mthd.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(new LabelComp("debug writer \"sym\" file", D_sym=new JTextField()));
    D_sym.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(new LabelComp("debug writer \"rel\" file", D_rel=new JTextField()));
    D_rel.addActionListener(defaultOptionChangedListener);
    tmpPanel.add(new JPanel());
    add(tmpPanel, BorderLayout.NORTH);
    tmpPanel=new JPanel();
    tmpPanel.setLayout(new GridLayout(0, 4, 2, 2));
    tmpPanel.add(new ThreeComp(new LabelComp("compile files", new JScrollPane(files=new JList(filesList=new DefaultListModel()))),
        fileAdd=new JButton("add"), fileDel=new JButton("del")));
    filesList.addElement("a"); filesList.addElement("a"); filesList.addElement("a"); filesList.addElement("a"); filesList.addElement("a");//insert dummy entries for later pack-call
    files.setVisibleRowCount(5);
    fileAdd.addActionListener(defaultAddDelListener);
    fileDel.addActionListener(defaultAddDelListener);
    tmpPanel.add(new ThreeComp(new LabelComp("directories", new JScrollPane(dirs=new JList(dirsList=new DefaultListModel()))),
        dirAdd=new JButton("add"), dirDel=new JButton("del")));
    dirs.setVisibleRowCount(5);
    dirAdd.addActionListener(defaultAddDelListener);
    dirDel.addActionListener(defaultAddDelListener);
    tmpPanel.add(new ThreeComp(new LabelComp("dir. non-rec.", new JScrollPane(dirsNonRec=new JList(dirsNonRecList=new DefaultListModel()))),
        dirNRAdd=new JButton("add"), dirNRDel=new JButton("del")));
    dirsNonRec.setVisibleRowCount(5);
    dirNRAdd.addActionListener(defaultAddDelListener);
    dirNRDel.addActionListener(defaultAddDelListener);
    tmpPanel.add(run=new JButton("run"));
    run.addActionListener(runListener);
    add(tmpPanel, BorderLayout.CENTER);
    add(new LabelComp("command line:", result=new JTextField()), BorderLayout.SOUTH);
    result.setEditable(false);
    pack();
    Dimension screen=Toolkit.getDefaultToolkit().getScreenSize();
    int width=getWidth();
    int height=getHeight();
    if (width>screen.getWidth()) width=(int)screen.getWidth();
    if (height>screen.getHeight()) height=(int)screen.getHeight();
    setSize(width, height);
    filesList.removeAllElements(); //clear dummy list
    setVisible(true);
    fileChooser=new JFileChooser();
    args=new Vector<String>();
  }
  
  private void appendOption(StringBuffer line, String option) {
    if (line.length()>0) line.append(' ');
    line.append(option);
    args.add(option);
  }
  
  private void appendOption(StringBuffer line, String opt1, String opt2) {
    if (line.length()>0) line.append(' ');
    line.append(opt1);
    line.append(' ');
    line.append(opt2);
    args.add(opt1);
    args.add(opt2);
  }
  
  private void appendOption(StringBuffer line, String opt1, String opt2, String opt3) {
    if (line.length()>0) line.append(' ');
    line.append(opt1);
    line.append(' ');
    line.append(opt2);
    line.append(' ');
    line.append(opt3);
    args.add(opt1);
    args.add(opt2);
    args.add(opt3);
  }
  
  private void triggerEvaluation(Object source) {
    //check for invalid combinations
    if (source==L_flat || source==l_streamline || source==e_embAddr || source==E_constInRam) {
      M_indirect.setSelected(false);
      m_dynaMem.setSelected(false);
    }
    else if (source==M_indirect || source==m_dynaMem) {
      L_flat.setSelected(false);
      l_streamline.setSelected(false);
      e_embAddr.setText("");
      E_constInRam.setSelected(false);
    }
    else if (source==h_heapAssign) c_allAssign.setSelected(false);
    else if (source==c_allAssign) h_heapAssign.setSelected(false);
    else if (source==B_skipBoundCheck) b_rteBound.setSelected(false);
    else if (source==b_rteBound) B_skipBoundCheck.setSelected(false);
    //build new command line
    args.clear();
    StringBuffer line=new StringBuffer();
    String text;
    if (v_verbose.isSelected()) appendOption(line, "-v");
    if (V_verboseTiming.isSelected()) appendOption(line, "-V");
    if (L_flat.isSelected()) appendOption(line, "-L");
    if (l_streamline.isSelected()) appendOption(line, "-l");
    if (M_indirect.isSelected()) appendOption(line, "-M");
    if (m_dynaMem.isSelected()) appendOption(line, "-m");
    if ((text=i_options.getText()).length()>0) appendOption(line, "-i", text);
    if ((text=H_listFiles.getText()).length()>0) appendOption(line, "-H", text);
    if (h_heapAssign.isSelected()) appendOption(line, "-h");
    if (c_allAssign.isSelected()) appendOption(line, "-c");
    if (w_alternateNew.isSelected()) appendOption(line, "-w");
    if (x_stackExtreme.isSelected()) appendOption(line, "-x");
    if (C_skipArrayCheck.isSelected()) appendOption(line, "-C");
    if (B_skipBoundCheck.isSelected()) appendOption(line, "-B");
    if (b_rteBound.isSelected()) appendOption(line, "-b");
    if (n_rteNull.isSelected()) appendOption(line, "-n");
    if (X_explicitConv.isSelected()) appendOption(line, "-X");
    if (d_mthdDebug.isSelected()) appendOption(line, "-n");
    if (g_genAllUnits.isSelected()) appendOption(line, "-g");
    if (G_genAllMthds.isSelected()) appendOption(line, "-G");
    if (j_genIntfParents.isSelected()) appendOption(line, "-j");
    if (q_relManager.isSelected()) appendOption(line, "-q");
    if (Q_srcHint.isSelected()) appendOption(line, "-Q");
    if (W_printCode.isSelected()) appendOption(line, "-W");
    if (F_profiler.isSelected()) appendOption(line, "-F");
    if (f_throwWithoutCatch.isSelected()) appendOption(line, "-f");
    if (y_byteString.isSelected()) appendOption(line, "-y");
    if (k_skipInlineMthd.isSelected()) appendOption(line, "-k");
    if (Y_nativeFloatInComp.isSelected()) appendOption(line, "-Y");
    if ((text=N_startMthd.getText()).length()>0) appendOption(line, "-N", text);
    if ((text=I_maxInlLev.getText()).length()>0) appendOption(line, "-I", text);
    if ((text=S_autoInlStmts.getText()).length()>0) appendOption(line, "-S", text);
    if ((text=K_allocMask.getText()).length()>0) appendOption(line, "-K", text);
    if ((text=s_imgSize.getText()).length()>0) appendOption(line, "-s", text);
    if (!(text=(String)t_arch.getSelectedItem()).equals("ia32")) appendOption(line, "-t", text);
    if ((text=T_archOpt.getText()).length()>0) appendOption(line, "-T", text);
    if ((text=a_addr.getText()).length()>0) appendOption(line, "-a", text);
    if ((text=Z_reloc.getText()).length()>0) appendOption(line, "-Z", text);
    if ((text=e_embAddr.getText()).length()>0) appendOption(line, "-e", text);
    if (E_constInRam.isSelected()) appendOption(line, "-E");
    if ((text=P_specialHeader.getText()).length()>0) appendOption(line, "-P", text);
    if ((text=p_namePrefix.getText()).length()>0) appendOption(line, "-p", text);
    if (!(text=(String)o_out.getSelectedItem()).equals("raw")) appendOption(line, "-o", text);
    if ((text=O_outPar.getText()).length()>0) appendOption(line, "-O", text);
    if (!(text=(String)u_sym.getSelectedItem()).equals("none")) appendOption(line, "-u", text);
    if ((text=U_symPar.getText()).length()>0) appendOption(line, "-U", text);
    if ((text=D_code.getText()).length()>0) appendOption(line, "-D", "code", text);
    if ((text=D_gcc.getText()).length()>0) appendOption(line, "-D", "gcc", text);
    if ((text=D_mthd.getText()).length()>0) appendOption(line, "-D", "mthd", text);
    if ((text=D_sym.getText()).length()>0) appendOption(line, "-D", "sym", text);
    if ((text=D_rel.getText()).length()>0) appendOption(line, "-D", "rel", text);
    Enumeration<?> elements=filesList.elements();
    while (elements.hasMoreElements()) appendOption(line, (String)elements.nextElement());
    elements=dirsList.elements();
    while (elements.hasMoreElements()) appendOption(line, (String)elements.nextElement());
    elements=dirsNonRecList.elements();
    while (elements.hasMoreElements()) appendOption(line, (String)elements.nextElement()+":");
    result.setText(line.toString());
  }
  
  private void tryDelItems(JList list) {
    DefaultListModel model=(DefaultListModel)list.getModel();
    for (Object o: list.getSelectedValues()) model.removeElement(o); 
  }
  
  private void tryAddItem(JList list, boolean filesNotDirs) {
    if (filesNotDirs) fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
    else fileChooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
    if (fileChooser.showOpenDialog(this)==JFileChooser.APPROVE_OPTION)
      ((DefaultListModel)list.getModel()).addElement(fileChooser.getSelectedFile().getAbsolutePath());
  }
  
  private void triggerAddDel(Object source) {
    if (source==fileAdd) tryAddItem(files, true);
    else if (source==fileDel) tryDelItems(files);
    else if (source==dirAdd) tryAddItem(dirs, false);
    else if (source==dirDel) tryDelItems(dirs);
    else if (source==dirNRAdd) tryAddItem(dirsNonRec, false);
    else if (source==dirNRDel) tryDelItems(dirsNonRec);
    else {
      JOptionPane.showMessageDialog(this, "internal SCG error");
      return;
    }
    triggerEvaluation(source);
  }
  
  private void doCompile() {
    setEnabled(false);
    final String[] argv=new String[args.size()];
    final String argString=result.getText();
    args.copyInto(argv);
    SwingUtilities.invokeLater(new Runnable() {
      public void run() {
        final JTextArea text;
        JFrame myWindow=new JFrame("SJC output");
        myWindow.add(new JScrollPane(text=new JTextArea()));
        text.append("calling SJC with parameters:\n");
        text.append(argString+"\n\n");
        text.setWrapStyleWord(false);
        text.setEditable(false);
        myWindow.setSize(400, 500);
        myWindow.setVisible(true);
        OutputStream textOut=new OutputStream() {
          public void write(int b) throws IOException {
            text.append((new Character((char)b)).toString());
          }
        };
        int res;
        Context ctx=new Context(new SunOS(textOut));
        symbols.SymbolFactory.preparedReflectionSymbols=new osio.sun.ReflectionSymbols();
        if ((res=ctx.compile(argv, "vJRE"))==0) ctx.writeSymInfo();
        else ctx.out.println("\ncompiler result: "+res);
      }
    });
    setEnabled(true);
  }
}
